package ij.gui;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import ij.*;

public class ImageCanvas extends Canvas implements MouseListener, MouseMotionListener, Cloneable {

	private ImagePlus imp;
	private ImageJ ij;
	private int imageWidth, imageHeight;
	private double magnification;
	private Rectangle srcRect;
	private int dstWidth, dstHeight;
	
	private int xMouseStart = 0;
	private int yMouseStart = 0;
	private int xSrcStart = 0;
	private int ySrcStart = 0;
	private int xMouse = -1;
	private int yMouse = -1;
	public static boolean bLive;  // Added by David Swartz for VPP plugin
	
	public ImageCanvas(ImagePlus imp) {
		this.imp = imp;
		ij = Info.getImageJ();
		int width = imp.getWidth();
		int height = imp.getHeight();
		setDrawingSize(width, height);
		imageWidth = width;
		imageHeight = height;
		srcRect = new Rectangle(0, 0, imageWidth, imageHeight);
		magnification = 1.0;
 		addMouseListener(this);
 		addMouseMotionListener(this);
 		addKeyListener(ij);  // ImageJ handles keyboard shortcuts
		add(Menus.getPopupMenu());
	}
		
	public void setDrawingSize(int width, int height) {
	    dstWidth = width;
	    dstHeight = height;
		setSize(dstWidth, dstHeight);
	}
		
	public void update(Graphics g) {
		if (Info.debugMode) Info.write("Update: " + imp.getTitle());
		paint(g);
	}

	public void print(Graphics g) {
		g.drawImage(imp.getImage(), 0, 0, null);
	}
	
    public void paint(Graphics g) {
		if (Info.debugMode) Info.write("Paint: " + imp.getTitle());
		//Info.write(srcRect.width + " " + srcRect.height + " -> " + dstWidth + " " + dstHeight);
		try {imp.draw();}
		catch(OutOfMemoryError e) {Info.outOfMemory("Paint");}
    }
    
    public Dimension preferredSize() {
        return new Dimension(dstWidth, dstHeight);
    }

	/** Returns the current cursor location. */
	public Point getCursorLoc() {
		return new Point(xMouse, yMouse);
	}

	/** Sets the cursor based on the current tool and cursor location. */
	public void setCursor(int x, int y) {
		if (x!=-1) xMouse = x;
		if (y!=-1) yMouse = y;
		Roi roi = imp.getRoi();
		ImageWindow win = imp.getWindow();
		if (win==null)
			return;
		if (Info.spaceBarDown()) {
			win.setCursor(win.HAND_CURSOR);
			return;
		}
		switch (Toolbar.getToolId()) {
			case Toolbar.MAGNIFIER:
				win.setCursor(win.MOVE_CURSOR);
				break;
			case Toolbar.HAND:
				win.setCursor(win.HAND_CURSOR);
				break;
			default:  //selection tool
				if (roi!=null && roi.getState()!=roi.CONSTRUCTING && roi.contains(x, y))
					win.setCursor(win.DEFAULT_CURSOR);
				else
					win.setCursor(win.CROSSHAIR_CURSOR);
		}
	}
		
	/**Converts a screen x-coordinate to an offscreen x-coordinate.*/
	public int offScreenX(int x) {
		return srcRect.x + (int)(x/magnification);
	}
	
	/**Converts a screen y-coordinate to an offscreen y-coordinate.*/
	public int offScreenY(int y) {
		return srcRect.y + (int)(y/magnification);
	}
	
	/**Converts an offscreen x-coordinate to a screen x-coordinate.*/
	public int screenX(int x) {
		return  (int)((x-srcRect.x)*magnification);
	}
	
	/**Converts an offscreen y-coordinate to a screen y-coordinate.*/
	public int screenY(int y) {
		return  (int)((y-srcRect.y)*magnification);
	}
	
	
	public double getMagnification() {
		return magnification;
	}
		
	public void setMagnification(double magnification) {
		this.magnification = magnification;
		ImageWindow win = imp.getWindow();
		if (magnification>1.0)
			win.setTitle(imp.getTitle() + " (" + (int)magnification + ":1)");
		else if (magnification<1.0)
			win.setTitle(imp.getTitle() + " (1:" + (int)(1/magnification) + ")");
		else
			win.setTitle(imp.getTitle());
	}
		
	public Rectangle getSrcRect() {
		return srcRect;
	}
	
	/** Enlarge the canvas if the user enlarges the window. */
	void resizeCanvas(int width, int height) {
		if (srcRect.width<imageWidth || srcRect.height<imageHeight) {
			if (width>imageWidth*magnification)
				width = (int)(imageWidth*magnification);
			if (height>imageHeight*magnification)
				height = (int)(imageHeight*magnification);
			setDrawingSize(width, height);
			srcRect.width = (int)(dstWidth/magnification);
			srcRect.height = (int)(dstHeight/magnification);
			if ((srcRect.x+srcRect.width)>imageWidth)
				srcRect.x = imageWidth-srcRect.width;
			if ((srcRect.y+srcRect.height)>imageHeight)
				srcRect.y = imageHeight-srcRect.height;
			imp.draw();
			//Info.write("resize: " + magnification + " " + srcRect.x+" "+srcRect.y+" "+srcRect.width+" "+srcRect.height+" "+dstWidth + " " + dstHeight);
		}
	}

	/** Zooms in by making the window bigger. If we can't
	make it bigger, then make the srcRect smaller.*/
	public void zoomIn(int x, int y) {
		if (magnification>=32)
			return;
		double newMag = magnification*2;
		if (1.0/newMag==imp.getWindow().getOriginalScale()) {
			unzoom();
			return;
		}
		int newWidth = dstWidth*2;
		int newHeight = dstHeight*2;
		Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
		ImageWindow win = imp.getWindow();
		Point loc = win.getLocationOnScreen();
		if ((loc.x+50+newWidth)<screen.width && (loc.y+50+newHeight)<screen.height) {
			setDrawingSize(newWidth, newHeight);
			win.pack();
		}
		else {
			int w = (int)Math.round(dstWidth/newMag);
			if (w*newMag<dstWidth) w++;
			int h = (int)Math.round(dstHeight/newMag);
			if (h*newMag<dstHeight) h++;
			x = offScreenX(x);
			y = offScreenY(y);
			Rectangle r = new Rectangle(x-w/2, y-h/2, w, h);
			if (r.x<0) r.x = 0;
			if (r.y<0) r.y = 0;
			if (r.x+w>imageWidth) r.x = imageWidth-w;
			if (r.y+h>imageHeight) r.y = imageHeight-h;
			srcRect = r;
		}
		setMagnification(newMag);
		imp.draw();
	}
		
	/**Zooms out by making srcRect bigger. If we can't make
	it bigger, then make the window smaller.*/
	void zoomOut(int x, int y) {
		if (magnification<=0.03125)
			return;
		double newMag = magnification/2;
		if (newMag==1.0) {
			unzoom();
			return;
		}
		if (srcRect.width<=imageWidth/2) {
			int w = (int)Math.round(dstWidth/newMag);
			if (w*newMag<dstWidth) w++;
			int h = (int)Math.round(dstHeight/newMag);
			if (h*newMag<dstHeight) h++;
			x = offScreenX(x);
			y = offScreenY(y);
			Rectangle r = new Rectangle(x-w/2, y-h/2, w, h);
			if (r.x<0) r.x = 0;
			if (r.y<0) r.y = 0;
			if (r.x+w>imageWidth) r.x = imageWidth-w;
			if (r.y+h>imageHeight) r.y = imageHeight-h;
			srcRect = r;
		}
		else {
			srcRect = new Rectangle(0, 0, imageWidth, imageHeight);
			setDrawingSize((int)(imageWidth*newMag), (int)(imageHeight*newMag));
			//setDrawingSize(dstWidth/2, dstHeight/2);
			imp.getWindow().pack();
		}
		//Info.write(newMag + " " + srcRect.x+" "+srcRect.y+" "+srcRect.width+" "+srcRect.height+" "+dstWidth + " " + dstHeight);
		setMagnification(newMag);
		//Info.write(srcRect.x + " " + srcRect.width + " " + dstWidth);
		imp.draw();
	}

	void unzoom() {
		if (magnification==1.0)
			return;
		setMagnification(1.0);
		srcRect = new Rectangle(0, 0, imageWidth, imageHeight);
		ImageWindow win = imp.getWindow();
		int scale = win.getOriginalScale();
		setMagnification(1.0/scale);
		setDrawingSize(imageWidth/scale, imageHeight/scale);
		win.pack();
		imp.draw();
	}
		
	void scroll(int sx, int sy) {
		int x = xSrcStart + (int)(sx/magnification);  //convert to offscreen coordinates
		int y = ySrcStart + (int)(sy/magnification);
		int newx = xSrcStart + (xMouseStart-x);
		int newy = ySrcStart + (yMouseStart-y);
		if (newx<0) newx = 0;
		if (newy<0) newy = 0;
		if ((newx+srcRect.width)>imageWidth) newx = imageWidth-srcRect.width;
		if ((newy+srcRect.height)>imageHeight) newy = imageHeight-srcRect.height;
		srcRect.x = newx;
		srcRect.y = newy;
		imp.draw();
	}	
	
	/** Draws the image using current magnification. */
	public void drawImage(Graphics g) {
		g.drawImage(imp.getImage(), 0, 0, (int)(srcRect.width*magnification), (int)(srcRect.height*magnification),
			srcRect.x, srcRect.y, srcRect.x+srcRect.width, srcRect.y+srcRect.height, null);
	}
	
	void setDrawingColor(int ox, int oy) {
		float[] v = imp.getPixel(ox, oy);
		Toolbar toolbar = Toolbar.getInstance();
		if (imp.getType()==imp.GRAY8) {
			int value = (int)v[0];
			if (imp.isInvertedLut())
				toolbar.setColor(new Color(255-value, 255-value, 255-value));
			else
				toolbar.setColor(new Color(value, value, value));
		} else
			toolbar.setColor(new Color((int)v[0], (int)v[1], (int)v[2]));
		Color c = toolbar.getColor();
		imp.setColor(c);
		Info.showStatus("Color set to ("+c.getRed()+", "+c.getGreen()+", "+c.getBlue()+")");
	}

	public void mousePressed(MouseEvent e) {
		int toolID = Toolbar.getToolId();
		ImageWindow win = imp.getWindow();
		if ( win instanceof StackWindow && ((StackWindow)win).animating==true && toolID!=Toolbar.MAGNIFIER) {
			((StackWindow)win).animating = false;
			return;
		}
		
		int x = e.getX();
		int y = e.getY();
		int flags = e.getModifiers();
		if (Info.debugMode) Info.write("Mouse pressed: (" + x + "," + y + ")" + ij.modifiers(flags));
		if (toolID!=Toolbar.MAGNIFIER && (e.isPopupTrigger() || (flags & e.META_MASK)!=0)) {
			if (Info.debugMode) Info.write("show popup: " + (e.isPopupTrigger()?"true":"false"));
			Menus.getPopupMenu().show(this, x, y);
			return;
		}

		int ox = offScreenX(x);
		int oy = offScreenY(y);
		if (Info.spaceBarDown()) {
			// temporarily switch to "hand" tool of space bar down
			xMouseStart = ox; yMouseStart = oy;
			xSrcStart = srcRect.x; ySrcStart = srcRect.y;
			return;
		}
		if ((flags&Event.ALT_MASK)!=0 && toolID!=Toolbar.MAGNIFIER) {
			// temporarily switch to "picker" tool alt/option key down
			setDrawingColor(ox, oy);
			return;
		}
		switch (toolID) {
			case Toolbar.MAGNIFIER:
				if ((flags & (Event.ALT_MASK|Event.META_MASK|Event.CTRL_MASK))!=0)
					zoomOut(x, y);
				else
					zoomIn(x, y);
				break;
			case Toolbar.HAND:
				xMouseStart = ox;
				yMouseStart = oy;
				xSrcStart = srcRect.x;
				ySrcStart = srcRect.y;
				break;
			case Toolbar.DROPPER:
				setDrawingColor(ox, oy);
				break;
			case Toolbar.WAND:
				Roi roi = imp.getRoi();
				if (roi!=null && roi.contains(ox, oy)) {
					Rectangle r = roi.getBoundingRect();
					if (r.width==imageWidth && r.height==imageHeight)
						imp.killRoi();
					else {
						handleRoiMouseDown(x, y);
						return;
					}
				}
				if (imp.getType()==imp.GRAY16 || imp.getType()==imp.GRAY32) {
					Info.error("The wand tool does not work with 16 and 32-bit grayscale images.");
					return;
				}
				Wand w = new Wand(imp.getProcessor());
				w.autoOutline(ox, oy);
				if (w.npoints>0) {
					imp.killRoi();
					roi = new PolygonRoi(w.xpoints, w.ypoints, w.npoints, imp);
					imp.setRoi(roi);
				}
				break;
			default:  //selection tool
				handleRoiMouseDown(x, y);
		}
	}

	public void mouseExited(MouseEvent e) {
		ImageWindow win = imp.getWindow();
		if (win!=null) win.setCursor(win.DEFAULT_CURSOR);
		xMouse = -1; yMouse = -1;
		Info.showStatus("");
	}

	public void mouseDragged(MouseEvent e) {
		int x = e.getX();
		int y = e.getY();
		if (Toolbar.getToolId()==Toolbar.HAND || Info.spaceBarDown())
			scroll(x, y);
		else {
			Roi roi = imp.getRoi();
			if (roi != null)
				roi.handleMouseDrag(x, y, (e.getModifiers()&e.SHIFT_MASK)!=0);
		}
	}

	void handleRoiMouseDown(int sx, int sy) {
		int ox = offScreenX(sx);
		int oy = offScreenY(sy);
		Roi roi = imp.getRoi();
		if (roi!=null) {
			Rectangle r = roi.getBoundingRect();
			int type = roi.getType();
			if (type==roi.RECTANGLE && r.width==imp.getWidth() && r.height==imp.getHeight()) {
				imp.killRoi();
				return;
			}
			if (roi.contains(ox, oy)) {
				roi.handleMouseDown(sx, sy);
				return;
			}
			if ((roi.getType()==roi.POLYGON || roi.getType()==roi.POLYLINE)
			&& roi.getState()==roi.CONSTRUCTING)
				return;
		}
		imp.createNewRoi(ox,oy);
	}

	public void mouseReleased(MouseEvent e) {
		Roi roi = imp.getRoi();
		if (roi != null) {
			Rectangle r = roi.getBoundingRect();
			if ((r.width==0 || r.height==0)
			&& !(roi.getType()==roi.POLYGON || roi.getType()==roi.POLYLINE)
			//&& !(roi instanceof TextRoi)
			&& roi.getState()==roi.CONSTRUCTING)
				imp.killRoi();
			else
				roi.handleMouseUp(e.getX(), e.getY());
		}
		Info.setKeyDown(KeyEvent.VK_UNDEFINED);
	}

    private String showValue(int ox, int oy) {
    	float[] v = imp.getPixel(ox, oy);
		switch (imp.getType()) {
			case imp.GRAY8: case imp.GRAY16: case imp.RGBA: case imp.HSB:
    			return(", value=" + (int)v[0]);
			case imp.GRAY32:
    			return(", value=" + v[0]);
			case imp.COLOR_256:
    			return(", index=" + (int)v[3] + ", value=" + (int)v[0] + "," + (int)v[1] + "," + (int)v[2]);
			case imp.COLOR_RGB:
    			return(", value=" + (int)v[0] + "," + (int)v[1] + "," + (int)v[2]);
    		default: return("");
		}
    }
        
	public void mouseMoved(MouseEvent e) {
		int ox = offScreenX(e.getX());
		int oy = offScreenY(e.getY());
		//if (Info.debugMode) Info.write(e.getX() + " " + e.getY() + " " + ox + " " + oy);
		setCursor(ox, oy);
		Roi roi = imp.getRoi();
		if (roi!=null && (roi.getType()==roi.POLYGON || roi.getType()==roi.POLYLINE) 
		&& roi.getState()==roi.CONSTRUCTING) {
			PolygonRoi pRoi = (PolygonRoi)roi;
			pRoi.handleMouseMove(ox, oy);
		}
		else {
			ImageWindow win = imp.getWindow();
			if (win instanceof HistogramWindow)
				((HistogramWindow)win).updateValueAndCount(ox, oy);
			else if (win instanceof PlotWindow)
				((PlotWindow)win).updateCursorCoordinates(ox, oy);
			Info.showStatus("Location = (" + ox + "," + oy + ")" + showValue(ox, oy));
		}
	}
	
	public void mouseClicked(MouseEvent e) { bLive = false; }  // Changed by David Swartz for VPP plugin
 	public void mouseEntered(MouseEvent e) {}

}